/*
 * connecting_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "connecting_screen.hpp"
#include "display.hpp"
#include "loading_screen.hpp"
#include "title_screen.hpp"
#include "x_centre.hpp"

#ifdef _WIN32_WCE
#include "wce_serial.hpp"
#else
#include "enet_streams/enet_controller.hpp"
#include "enet_streams/enet_stream.hpp"
#endif

#include "boost/scoped_ptr.hpp"
using namespace boost;

namespace {
    ConnectingScreen *conscr = 0;
}

void SetupConnectingScreen(const UsefulStuff &u)
{
    ASSERT(!conscr);
    conscr = new ConnectingScreen(u);
    conscr->render();
    u.app2d->addObserver(conscr);
    u.app2d->setUpdateIntervalActive(50);
    u.app2d->setUpdateIntervalInactive(50);
    u.app2d->setRenderFast(false);
}

void StopConnectingScreen(const UsefulStuff &u)
{
    if (!conscr) return;
    u.app2d->rmObserver(conscr);
    delete conscr;
    conscr = 0;
}

ConnectingScreen::ConnectingScreen(const UsefulStuff &u)
    : us(u), awaiting_input(false)
{
#ifdef _WIN32_WCE
    // Open the COM port
    if (us.server) {
        wce_str.reset(new WceSerialStream(CfgInt("in_com")));
    } else {
        wce_str.reset(new WceSerialStream(CfgInt("out_com")));
    }
    us.remote = wce_str;
#else
    // Start up enet controller, and start up server if appropriate
    us.enet_controller.reset(new EnetController);
    if (us.server) {
        us.enet_controller->startServer(CfgInt("port_number"));
    } else {
        awaiting_input = true;
    }
#endif
}

void ConnectingScreen::render()
{
    us.display->setFontSize(CfgInt("font_size"));
    const int text_height = us.display->getTextHeight();
    us.display->begin();
    
    if (us.server) {
        XCentre(*us.display, text_height + 15, "WAITING FOR");
        XCentre(*us.display, 2*text_height + 20, "CONNECTION");

    } else {

        if (awaiting_input) {
            XCentre(*us.display, text_height + 15, "Enter IP address to connect to:");
            XCentre(*us.display, 2*text_height + 20, (ip_address + '|').c_str());
        } else {
            XCentre(*us.display, text_height + 15, "CONNECTING TO");
            XCentre(*us.display, 2*text_height + 20, ip_address.c_str());
        }
    }

    us.display->end();
}

void ConnectingScreen::update()
{
    bool done = false;
#ifdef _WIN32_WCE
    done = wce_str->tryConnect();
#else
    if (us.server) {
        // Listen for clients connecting to the server
        if (clients.empty()) {
            us.enet_controller->getClients(clients);
            us.enet_controller->processMessages(100); // allow enet to block for a bit.
        }
        if (!clients.empty()) {
            // Somebody has connected -- we are done. 
            us.remote = clients.front();
            done = true;
        }
    } else {

        if (!awaiting_input) {
            // Connect to the server; time out if the server does not respond within 5 sec.
            const int port_number(CfgInt("port_number"));
            us.remote = us.enet_controller->connectToServer(ip_address.c_str(), port_number, 5000);
            if (!us.remote) {
                // Go back to the title screen.
                SetupTitleScreen(us);
                StopConnectingScreen(us);
            } else {
                done = true;
            }
        }
    }
#endif

    if (done) {
        // Enable exceptions on the remote stream. This should catch unexpected EOF errors.
        us.remote->clear();
        us.remote->exceptions(ios_base::badbit | ios_base::failbit | ios_base::eofbit);
        
        // Go to the loading screen next.
        SetupLoadingScreen(us);
        StopConnectingScreen(us);
    }
}

void ConnectingScreen::key(Keycode key)
{
    if (key <= 255 && key >= 32) {
        ip_address += tolower(char(key));
        render();
    } else if (key >= KC_NUM_0 && key <= KC_NUM_9) {
        ip_address += ('0' + (key - KC_NUM_0));
    } else if (key == KC_NUM_DOT) {
        ip_address += '.';
    } else if (key == KC_BACKSPACE) {
        if (!ip_address.empty()) {
            ip_address = ip_address.substr(0, ip_address.length()-1);
        }
    } else if (key == KC_ESCAPE) {
        // go back to title screen
        SetupTitleScreen(us);
        StopConnectingScreen(us);
    } else if (key == KC_RETURN || key == KC_NUM_ENTER) {
        awaiting_input = false;
        render();
    }
}
