/*
 * draw.cpp
 *
 * The various in-game drawing routines
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "colour_change.hpp"
#include "draw.hpp"
#include "display.hpp"
#include "graphic.hpp"
#include "round.hpp"

namespace {
    const int NCOLS = 4;
    unsigned char COLS[NCOLS][3] = {
        { 255,0,0 },
        { 255,255,255 },
        { 255,0,0 },
        { 0,0,0 }
    };
}

//
// drawBackpackEntry
//

void DrawUI::drawBackpackEntry(Display &display, int left, int bottom, int width, int height,
                               int num_slots, int spacing,
                               int slot, const Graphic *gfx, const Graphic *overdraw_, int no_carried, int no_max)
{
    if (!gfx) return;
    if (no_carried == 0) return;
    if (spacing < 1) spacing = 1;

    const Graphic *overdraw = overdraw_ ? overdraw_ : gfx;
    
    display.clearClipRectangle();

    // assumption: all item gfx, including gems, are the same width (although
    // not necessarily the same height).
    const int width_of_each = width / num_slots;
    const float scale_factor = float(width_of_each) / float(GetWidth(*gfx));
    const int rescaled_height = Round(float(GetHeight(*gfx)) * scale_factor);
    const int rescaled_overdraw_height = Round(float(GetHeight(*overdraw)) * scale_factor);
    
    // Slots: 11, 12, ... 19 are Normal Items
    // 20 = Lock Picks
    // 21, 22, 23 = Keys.
    // 30 = Gems

    if (slot > 10 && slot < 20) {
        // Normal items are drawn in a column format, starting from the 2nd slot
        // (the 1st being reserved for keys).
        // Vertical spacing is chosen such that no_max items fit exactly into the height
        // of the inventory area. (If no_max==0, then space them by the supplied spacing value.)

        const int x = left + width_of_each*(slot-10);
        const int ht = rescaled_height;

        if (no_max != 0) {
            const int num = height - ht;
            if (no_carried > no_max) no_max = no_carried;
            for (int j = no_max - no_carried + 1; j <= no_max; ++j) {
                display.drawGraphic(x, num*j/no_max, *gfx, width_of_each, rescaled_height);
            }
        } else {
            // use supplied spacing; extra ones are just dropped.
            const int jmax = (height - ht) / spacing;
            for (int j = min(no_carried-1, jmax); j>=0; --j) {
                display.drawGraphic(x, bottom + height - ht - j*spacing, j ? *overdraw : *gfx, width_of_each,
                                    j ? rescaled_overdraw_height : rescaled_height);
            }
        }
        
    } else if (slot >= 20 && slot <= 29) {
        // Keys are drawn downwards from the top, in the leftmost slot.
        // NB no_carried and no_max are ignored for keys and lockpicks.
        const int x = left;
        const int ht = rescaled_height;
        const int y = bottom + height - ht*(slot-20);
        display.drawGraphic(x, y, *gfx, width_of_each, rescaled_height);

    } else if (slot == 30) {
        // Gems are drawn UNDERNEATH the normal inventory slots, in a right to left order.
        int x = left + width_of_each*(num_slots-1);
        const int y = bottom - rescaled_height;
        for (int i = 0; i < no_carried; ++i) {
            display.drawGraphic(x, y, *gfx, width_of_each, rescaled_height);
            x -= width_of_each;
        }
    }
}


//
// drawMessage
//

void DrawUI::drawMessage(Display &display, int bl_x, int bl_y, int tr_x, int tr_y,
                         int sq_rel_x, int sq_rel_y, int w, int h, int pixels_per_square, const string &message)
{
    const int room_bl_x = (tr_x - bl_x - w*pixels_per_square)/2 + bl_x;
    const int room_bl_y = (tr_y - bl_y - h*pixels_per_square)/2 + bl_y;
    const int sq_bl_x = sq_rel_x*pixels_per_square + room_bl_x;
    const int sq_bl_y = sq_rel_y*pixels_per_square + room_bl_y;
    int text_width, text_height;
    display.getTextSize(message.c_str(), text_width, text_height);
    const int tx = sq_bl_x + (pixels_per_square/2) - (text_width/2);
    const int ty = sq_bl_y + pixels_per_square*11/10;

    const int text_margin_x = text_height / 2;
    const int text_margin_y = text_height / 10;
    
    display.setClipRectangle(bl_x, bl_y, tr_x, tr_y);
    display.drawBlendedRectangle(tx - text_margin_x, ty - text_margin_y,
                                 text_width + text_margin_x*2, text_height + text_margin_y*2,
                                 0, 0, 0, CfgInt("msg_alpha"));
    display.drawText(tx, ty, message.c_str());
}


//
// drawMenu
// NOTE: we assume that graphics are square and all the same size.
//

void DrawUI::drawMenu(Display &display, int bl_x, int bl_y, int size,
                      const Graphic * gfx_centre, const Graphic * gfx_empty,
                      const Graphic * gfx_highlight, int gvt,
                      const Graphic * gfx_north, const Graphic * gfx_east,
                      const Graphic * gfx_south, const Graphic * gfx_west,
                      bool highlight, MapDirection highlight_dir)
{
    display.clearClipRectangle();
    if (gfx_centre) {
        const int w = size;
        const int h = size;
        display.drawGraphic(bl_x + w, bl_y + h, *gfx_centre, w, h);
        drawMenuItem(display, gfx_north, gfx_empty, bl_x + w, bl_y + 2*h, w, h);
        drawMenuItem(display, gfx_west, gfx_empty, bl_x, bl_y + h, w, h);
        drawMenuItem(display, gfx_east, gfx_empty, bl_x + 2*w, bl_y + h, w, h);
        drawMenuItem(display, gfx_south, gfx_empty, bl_x + w, bl_y, w, h);
        if (gfx_highlight) {
            int xofs = 0, yofs = 0;
            if (highlight) {
                switch (highlight_dir) {
                case D_NORTH:
                    yofs = h;
                    break;
                case D_EAST:
                    xofs = w;
                    break;
                case D_SOUTH:
                    yofs = -h;
                    break;
                case D_WEST:
                    xofs = -w;
                    break;
                }
            }
            ColourChange cc;
            const int n = (gvt / CfgInt("mini_map_flash_time")) % NCOLS;
            cc.add(255, 0, 0, COLS[n][0], COLS[n][1], COLS[n][2]);
            display.drawGraphicWithCC(bl_x + xofs + w, bl_y + yofs + h, *gfx_highlight, w, h, cc);
        }
    }
}

void DrawUI::drawMenuItem(Display &disp, const Graphic *g, const Graphic *e, int x, int y, int w, int h)
{
    if (g) disp.drawGraphic(x,y,*g, w, h);
    else if (e) disp.drawGraphic(x,y,*e, w, h);
}



