/*
 * enet_controller.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "enet_controller.hpp"
#include "enet_stream.hpp"

#include <sstream>
using namespace std;

EnetController::EnetController()
{
    server = 0;
    client = 0;
    client_peer.peer = 0;

    // Initialize the ENet library
    if (enet_initialize() != 0) {
        throw EnetError("failed to initialize ENet library");
    }
}

EnetController::~EnetController()
{
    // Do a shutdown, this will forcibly close any open connections
    shutDown(0);
    
    // "Close" any remaining connection streams
    for (vector<PeerData>::iterator it = server_peers.begin(); it != server_peers.end();
    ++it) {
        it->stream->close();
    }
    if (client_peer.stream) client_peer.stream->close();
    
    // Destroy ENetHost objects (this should reset any open connections, too)
    if (server) enet_host_destroy(server);
    if (client) enet_host_destroy(client);

    // Close down the ENet library
    enet_deinitialize();
}


//
// server side
//

void EnetController::startServer(int port, int maxclients, int bw_incoming,
                                 int bw_outgoing)
{
    if (server || client) throw EnetError("can't start server, ENet already running");
    
    ENetAddress address;
    address.host = ENET_HOST_ANY;
    address.port = port;

    server = enet_host_create(&address, maxclients, bw_incoming, bw_outgoing);
    if (!server) {
        throw EnetError("Failed to create an ENet server host");
    }
}
    
void EnetController::getClients(vector<shared_ptr<EnetStream> > &result)
{
    // make sure we are up to date
    processMessages();

    // create a list of streams
    result.clear();
    for (vector<PeerData>::iterator it = server_peers.begin(); it != server_peers.end();
    ++it) {
        result.push_back(it->stream);
    }
}

string EnetController::getIP() const
{
    if (!server) return "(unknown)";
    const ENetAddress & addr = server->address;
    const unsigned char * ptr = (unsigned char*)(&addr.host);
    ostringstream str;
    str << int(ptr[0]) << "." << int(ptr[1]) << "." << int(ptr[2]) << "." << int(ptr[3]);
    return str.str();
}

//
// Client side
//

shared_ptr<EnetStream> EnetController::connectToServer(const char *address, int port,
    int timeout, int bw_incoming, int bw_outgoing)
{
    if (server || client) throw EnetError("can't start client, ENet already running");

    // Create 'client' object
    client = enet_host_create(0, 1, bw_incoming, bw_outgoing);
    if (!client) throw EnetError("Failed to create an ENet client host");

    // Create a 'peer' representing the server
    ENetAddress en_addr;
    enet_address_set_host(&en_addr, address);
    en_addr.port = port;
    ENetPeer *peer = enet_host_connect(client, &en_addr, 1);  // open 1 channel only
    if (!peer) throw EnetError("Failed to create ENet peer while connecting");

    // We must now wait for an ENET_EVENT_TYPE_CONNECT event
    ENetEvent event;
    if (enet_host_service(client, &event, timeout) > 0 &&
    event.type == ENET_EVENT_TYPE_CONNECT) {
        // Success
        client_peer.peer = peer;
        client_peer.stream.reset(new EnetStream(this, peer));
        return client_peer.stream;
    } else {
        // Connection failed. Reset the peer and destroy the client.
        enet_peer_reset(peer);
        enet_host_destroy(client);
        client = 0;
        return shared_ptr<EnetStream>();
    }
}


//
// processMessages
//

void EnetController::processMessages(int timeout, bool allow_conn)
{
    ENetHost *host = 0;
    if (client) host = client;
    if (server) host = server;
    if (!host) return;  // nothing to do

    ENetEvent event;
    while ((allow_conn || (server? !server_peers.empty() : bool(client))) &&
    enet_host_service(host, &event, timeout) > 0) {
        switch (event.type) {
            
        case ENET_EVENT_TYPE_CONNECT:
            ASSERT(server);
            // A new client has connected to the server
            // Create a new EnetStream representing the new client, and add it to the list
            // of connected clients.
            if (allow_conn) {
                PeerData p;
                p.peer = event.peer;
                p.stream.reset(new EnetStream(this, p.peer));
                server_peers.push_back(p);
            }
            break;

        case ENET_EVENT_TYPE_RECEIVE:
            // Packet received -- pass it on to the relevant streambuf.
            // (Note that we ignore the channelID.)
            {
                EnetStream *stream;
                if (client) {
                    // Message directed to a client
                    stream = client_peer.stream.get();
                } else if (server) {
                    // Message from a client -- find out which stream to send it to
                    vector<PeerData>::iterator it;
                    for (it = server_peers.begin(); it != server_peers.end(); ++it) {
                        if (it->peer == event.peer) {
                            stream = it->stream.get();
                            break;
                        }
                    }
                }
                if (stream) stream->receivePacket(event.packet);
            }
            break;
            
        case ENET_EVENT_TYPE_DISCONNECT:
            if (client) {
                // We disconnected from the server
                ASSERT(event.peer == client_peer.peer);
                client_peer.peer = 0;
                client_peer.stream->close();
                client_peer.stream.reset();

                // Destroy the client host too, since we won't need it any more
                enet_host_destroy(client);
                client = 0;

                // no more msgs to process
                return;
            } else {
                // A client disconnected from us
                vector<PeerData>::iterator it;
                for (it = server_peers.begin(); it != server_peers.end(); ++it) {
                    if (it->peer == event.peer) break;
                }
                ASSERT(it != server_peers.end());
                it->stream->close();
                server_peers.erase(it);
            }
            break;
        }
    }
}

void EnetController::shutDown(int timeout)
{
    if (!server && !client) return;  // nothing to do

    // Ask ENet to disconnect all peers
    if (server) {
        for (vector<PeerData>::iterator it = server_peers.begin(); it != server_peers.end();
        ++it) {
            enet_peer_disconnect(it->peer);
        }
    } else if (client) {
        ASSERT(client_peer.peer);
        enet_peer_disconnect(client_peer.peer);
    }

    // Wait for the disconnect messages to come in
    processMessages(timeout, false);

    // Make sure we reset any remaining peers
    if (server) {
        for (vector<PeerData>::iterator it = server_peers.begin(); it != server_peers.end();
        ++it) {
            enet_peer_reset(it->peer);
            it->stream->close();
        }
    } else if (client) {
        ASSERT(client_peer.peer);
        enet_peer_reset(client_peer.peer);
        client_peer.stream->close();
    }

    // Reset member variables.
    server = 0;
    server_peers.clear();
    client = 0;
    client_peer.peer = 0;
    client_peer.stream.reset();
}
