/*
 * in_game_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "client_config.hpp"
#include "game_client.hpp"
#include "game_server.hpp"
#include "in_game_screen.hpp"
#include "local_dungeon_view.hpp"
#include "mediator.hpp"
#include "menu_screen.hpp"
#include "misc/my_exceptions.hpp"
#include "display.hpp"
#include "remote_dungeon_view.hpp"
#include "server_config.hpp"
#include "task_manager.hpp"
#include "x_centre.hpp"

namespace {
    InGameScreen * igs = 0;
}

void SetupInGameScreen(const UsefulStuff &u, const MenuSelections &msel)
{
    ASSERT(!igs);  // be careful not to create more than one InGameScreen at a time!
    igs = new InGameScreen(u, msel);
    u.display->begin(); u.display->end();  // clear the screen
    u.app2d->addObserver(igs);
    const int ingame_update_interval = 1000 / CfgInt("fps");
    u.app2d->setUpdateIntervalActive(ingame_update_interval);
    u.app2d->setUpdateIntervalInactive(ingame_update_interval);
    u.app2d->setRenderFast(false);  // render after each update (only)
}

void StopInGameScreen(const UsefulStuff &u)
{
    if (!igs) return;
    u.app2d->rmObserver(igs);
    delete igs;
    igs = 0;
}

InGameScreen::InGameScreen(const UsefulStuff &u, const MenuSelections &msel)
    : us(u), first_update(false), sent_ready_yet(false), knights_spawned_yet(false),
      prev_update_time(0), stage(SETUP_STAGE)
{
    if (us.server) {
        // Server setup is done here (client setup is done in separate "setup stage")
        serverSetup(msel);
        stage = HANDSHAKE_STAGE;
    }
}

InGameScreen::~InGameScreen()
{
    // Upon destruction of InGameScreen (end of game) we can delete the GameClient,
    // GameServer and the DungeonViews.
    game_client.reset();
    game_server.reset();
    rdv1.reset();
    ldv2.reset();
    ldv1.reset();
}

namespace {
    // Helper routines to configure DungeonViews properly.
    
    // Note there are two versions, one based on a ServerConfig and one that sends/receives
    // the necessary information over the network (to clients).

    // [Maybe potion and skull renderers, and menu gfx, should be shared? (TODO)]

    void SetupLocalDungeonView(LocalDungeonView &ldv, ServerConfig &server_config)
    {
        ldv.setMiniMap(Mediator::instance().getMap()->getWidth(),
                       Mediator::instance().getMap()->getHeight());
        ldv.setPotionRenderer(&server_config.getPotionRenderer());
        ldv.setSkullRenderer(&server_config.getSkullRenderer());
        ldv.setMenuGfx(server_config.getMenuCentre(), server_config.getMenuEmpty(),
                       server_config.getMenuHighlight());
    }
    
    void SendRemoteDviewSetup(iostream &str, GameServer &game_server)
    {
        BinWriteUword(str, Mediator::instance().getMap()->getWidth());
        BinWriteUword(str, Mediator::instance().getMap()->getHeight());
    }
    
    void ReceiveRemoteDviewSetup(LocalDungeonView &dview, iostream &str,
                                 ClientConfig &client_config)
    {
        int w = BinReadUword(str);
        int h = BinReadUword(str);
        dview.setMiniMap(w, h);
        dview.setPotionRenderer(client_config.getPotionRenderer());
        dview.setSkullRenderer(client_config.getSkullRenderer());
        dview.setMenuGfx(client_config.getMenuCentre(), client_config.getMenuEmpty(),
                         client_config.getMenuHighlight());
    }
}

void InGameScreen::serverSetup(const MenuSelections &msel)
{
    // First: create DungeonViews
    if (us.remote) {
        // Server game
        ASSERT(us.server_config);
        ldv1.reset(new LocalDungeonView(*us.right_controller));
        rdv1.reset(new RemoteDungeonView(*us.remote, *us.server_config));
    } else {
        // Local game
        ldv1.reset(new LocalDungeonView(*us.left_controller));
        ldv2.reset(new LocalDungeonView(*us.right_controller));
    }

    // Create GameServer
    vector<DungeonView*> dviews;
    if (ldv1) dviews.push_back(ldv1.get());
    if (ldv2) dviews.push_back(ldv2.get());
    if (rdv1) dviews.push_back(rdv1.get());
    game_server.reset(new GameServer(2));  // 2 players
    game_server->initialize(*us.server_config, msel, dviews);

    // If we get to here, then the dungeon was successfully generated. Send out the 'M'
    // command to the remote (this closes the menu screen on the other side).
    if (us.remote) {
        us.remote->put('M');
        us.remote->flush();
    }
    
    // Do local dview setup
    if (ldv1) SetupLocalDungeonView(*ldv1, *us.server_config);
    if (ldv2) SetupLocalDungeonView(*ldv2, *us.server_config);

    // Send out remote dview information
    if (us.remote) {
        SendRemoteDviewSetup(*us.remote, *game_server);
        us.remote->flush();
    }
}

bool InGameScreen::tryClientSetup()
{
    // Create DungeonView if necessary
    if (!ldv1) {
        ldv1.reset(new LocalDungeonView(*us.right_controller));
    }

    // Create GameClient if necessary
    if (!game_client) {
        // (note: GameClient ctor does not actually read anything from remote)
        game_client.reset(new GameClient(*us.remote, *ldv1, *us.client_config));
    }
    
    // Receive the remote dview information.
    // (This fails if there is no data available to read)
    if (us.remote->rdbuf()->in_avail() == 0) return false;
    ReceiveRemoteDviewSetup(*ldv1, *us.remote, *us.client_config);

    // We are done!
    return true;
}

void InGameScreen::render()
{
    static const int ref_vp_width = CfgInt("dpy_viewport_width");
    static const int ref_vp_height = CfgInt("dpy_viewport_height");
    static const int ref_gutter = CfgInt("dpy_gutter");
    
    us.display->begin();
    if (!game_client && !game_server) {
        XCentre(*us.display, us.display->getTextHeight() + 15, "STARTING GAME");
    } else {
        const int gvt = game_client? game_client->getGVT()
            : Mediator::instance().getGVT();

        // Have to compute appropriate scaled rectangles
        // NOTE: For split screen mode, assume 2 viewports, laid out horizontally.
        // For networked mode, we assume one viewport only.
        
        const int dpy_width = us.display->getWidth();
        const int dpy_height = us.display->getHeight();
        const int ref_width = ldv2 ? ref_vp_width * 2 + ref_gutter : ref_vp_width;
        const int ref_height = ref_vp_height;

        const GfxResizer &resizer = *us.display->getGfxResizer();
        float dummy;

        const float ideal_x_scale = float(dpy_width) / float(ref_width);
        const float ideal_y_scale = float(dpy_height) / float(ref_height);
        
        float x_scale, y_scale;
        resizer.roundScaleFactor(ideal_x_scale, x_scale, dummy);
        resizer.roundScaleFactor(ideal_y_scale, y_scale, dummy);
        const float scale = min(x_scale, y_scale);
        
        const int x_bias = int(ref_width * (ideal_x_scale - scale)) / 2;
        const int y_bias = int(ref_height * (ideal_y_scale - scale)) / 2;
        const int scaled_vp_width = int(ref_vp_width * scale);
        const int scaled_gutter = int(ref_gutter * scale);
        
        if (ldv1) ldv1->draw(*us.display, scale, x_bias, y_bias, gvt);
        if (ldv2) ldv2->draw(*us.display, scale, x_bias + scaled_vp_width + scaled_gutter, y_bias, gvt);
    }
    us.display->end();
}

void InGameScreen::update()
{
    int c;
    int amount;
    int ms;
    
    switch (stage) {
    case SETUP_STAGE:
        ASSERT(!us.server);
        if (tryClientSetup()) {
            stage = HANDSHAKE_STAGE;
        }
        break;
        
    case HANDSHAKE_STAGE:
        // At this point we do a crude synchronization between the two machines. This is
        // mainly to prevent the server from starting before the client is ready.
        
        if (!us.remote) {
            // Don't need this for local games:
            stage = INGAME_STAGE;
            break;
        }
        
        // Send 'ready'
        if (!sent_ready_yet) {
            us.remote->put('R');
            us.remote->flush();
            sent_ready_yet = true;
        }
        
        // Wait for 'ready'
        if (us.remote->rdbuf()->in_avail() == 0) break;
        c = us.remote->get();
        if (c != int('R')) {
            throw UnexpectedError("Handshake failure");
        }
        
        // OK, both sides are now roughly synchronized (up to lag delays etc).
        // Proceed to In-Game stage.
        stage = INGAME_STAGE;
        break;

    case INGAME_STAGE:
        {
            // Spawn knights if necessary
            if (!knights_spawned_yet && game_server) {
                game_server->spawnPlayers();
                knights_spawned_yet = true;
            }
            
            // Check for 'quit' commands on the client
            if (game_client && game_client->isGameQuit()) {
                stage = GAMEOVER_STAGE;
                break;
            }
            
            // Work out how much time has passed since last update
            ms = us.app2d->getMilliseconds();
            if (first_update) {
                amount = 1;
                first_update = false;
                prev_update_time = ms;
            } else {
                amount = ms - prev_update_time;
                prev_update_time = ms;
                if (amount <= 0) return;
            }
            
            amount *= CfgInt("ticks_per_second");
            amount /= 1000;
            
            const int max_update = CfgInt("max_update");
            if (amount > max_update) {
                amount = max_update; // prevent too big jumps!
            }
            
            // Now do an actual update
             
            // NOTE: The in-game network protocol is actually quite simple:
            // (i) Server sends "dungeonview commands" to client (these are sent from RemoteDungeonView on
            // server side; on client side they are received in GameClient::update / GameClient::receiveCommands
            // and then forwarded to the client's LocalDungeonView).
            // (ii) Client sends "controls" to server - the controls are just 1 byte IDs. (These originate
            // in the client's LocalDungeonView, are transmitted from GameClient::update on client side,
            // and received in RemoteDungeonView::readController on server side.) Alternatively client may send
            // the byte 255 to acknowledge that the game has ended.
            
            if (game_client) {
                game_client->update(amount);
            } else if (game_server) {
                game_server->update(Mediator::instance().getGVT() + amount);
            }
        }
        break;

    case GAMEOVER_STAGE:
        // If this is the server, we must now wait for the client to confirm that
        // it has received the "game over" message (no. 255; see also
        // RemoteDungeonView::readController and GameClient::sendGameOverMsg).
        if (game_server && us.remote) {
            if (us.remote->rdbuf()->in_avail() == 0) break;
            c = us.remote->get();
            if (c != 255) {
                // discard non-255 characters as there may still be incoming commands pending.
                break;
            }
        }

        // At this point we must go back to the menu screen.
        SetupMenuScreen(us);
        StopInGameScreen(us);
        break;
    }
}

void InGameScreen::mouseDown(int x, int y, bool rmb)
{
    if (stage != INGAME_STAGE) return;
    if (game_client && game_client->isGameOver()
    || game_server && Mediator::instance().getWinningPlayer()) {
        stage = GAMEOVER_STAGE; // this will reset everything and go back to menu
    }
}

void InGameScreen::key(Keycode key)
{
    if (game_server && stage == INGAME_STAGE) {
        bool quit = false;
#ifdef _WIN32_WCE
        if (key == KC_RIGHT) {
            quit = true;
        }
#else
        if (key == KC_ESCAPE) {
            quit = true;
        }
#endif
        if (quit) {
            // tell server to send out QUIT messages to clients
            Mediator::instance().quitGame();
            // for our part, we go directly to GAMEOVER_STAGE...
            stage = GAMEOVER_STAGE;
        }
    }
}
