/*
 * local_dungeon_view.hpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef LOCAL_DUNGEON_VIEW_HPP
#define LOCAL_DUNGEON_VIEW_HPP

#include "dungeon_view.hpp"
#include "entity_map.hpp"
#include "graphic_element.hpp"
#include "mini_map.hpp"
#include "potion_renderer.hpp"
#include "skull_renderer.hpp"

#include <deque>
#include <list>
#include <map>
#include <queue>
using namespace std;

class Control;
class Controller;
class Display;
class ItemType;
class PotionRenderer;
class SkullRenderer;
class Tile;

class LocalDungeonView : public DungeonView {
public:
    explicit LocalDungeonView(const Controller &ctrlr);
    void setMiniMap(int w, int h);
    void setSkullRenderer(const SkullRenderer *s) { skull_renderer = s; }
    void setPotionRenderer(const PotionRenderer *p) { potion_renderer = p; }
    void setMenuGfx(const Graphic *centre, const Graphic *empty, const Graphic *highlight) {
        menu_gfx_centre = centre; menu_gfx_empty = empty; menu_gfx_highlight = highlight;
    }
    
    // rooms and mapping
    virtual void setCurrentRoom(int r, int w, int h);
    virtual int getCurrentRoom() const;
    virtual void mapCurrentRoom(const MapCoord &bl_pos);
    virtual void mapWall(const MapCoord &mc);
    virtual void mapKnightLocation(int id, const MapCoord &pos);
    virtual void mapItemLocation(const MapCoord &pos, bool on);
    virtual void wipeMap();
    virtual bool isSquareMapped(const MapCoord &mc) const;
    
    // entities
    virtual void addEntity(unsigned short int id, int x, int y, MapHeight height, MapDirection facing, 
                           const Anim * anm, const Overlay * ovr, int af, int atz_diff,
                           bool ainvuln,
                           int cur_ofs, MotionType motion_type, int motion_time_remaining);
    virtual void rmEntity(unsigned short int id);
    virtual void repositionEntity(unsigned short int id, int new_x, int new_y);
    virtual void moveEntity(unsigned short int id, MotionType mt, int motion_duration, bool missile_mode);
    virtual void flipEntityMotion(unsigned short int id, int initial_delay, int motion_duration);
    virtual void setAnimData(unsigned short int id, const Anim *, const Overlay *, int af,
                             int atz_diff, bool ainvuln, bool currently_moving);
    virtual void setFacing(unsigned short int id, MapDirection new_facing);
    
    // tiles
    virtual void clearTiles(int x, int y);
    virtual void setTile(int x, int y, int depth, const Graphic *gfx,
                         shared_ptr<const ColourChange> cc, MiniMapColour col);

    // items
    virtual void setItem(int x, int y, const Graphic *gfx);

    // "icons" (used for blood splats)
    virtual void placeIcon(int x, int y, const Graphic *, int dur);

    // backpack, health, skulls
    virtual void setBackpack(int slot, const Graphic *gfx, const Graphic *overdraw, int no_carried, int no_max); 
    virtual void addSkull();
    virtual void setHealth(int h);
    virtual void setPotionMagic(PotionMagic pm, bool pi);
    
    // messages
    virtual void flashMessage(const string &s, int ntimes);
    virtual void cancelContinuousMessages();
    virtual void addContinuousMessage(const string &s);

    // flash screen
    virtual void flashScreen(int delay);

    // sound
    virtual void playSound(const string &name, int frequency);
    
    // winning and losing
    virtual void winGame() { won = true; }
    virtual void loseGame() { lost = true; }
    virtual void quitGame() { } 
    
    // controls and menu
    virtual void setAvailableControls(const vector<pair<const Control *,bool> > &controls);
    virtual void setMenuHighlight(const Control *ctrl);
    virtual const Control * readController();

    // drawing
    // gfx drawn at scale*(x,y) + (origin_x, origin_y) in display coords, where (x,y) are the reference coords.
    void draw(Display &, float scale, int origin_x, int origin_y, int new_gvt);

    // add cheat controls
    // (this is done by GameServer at setup time)
    virtual void addCheatControls(const vector<const Control*> *c) { cheat_controls = c; }
    
private:
    void setMenuControl(MapDirection, const Control *, const Control *);
    bool menuNull() const { return menu_null==M_NULL; }
    
private:
    // gvt (needed to keep track of entity motions)
    int gvt;
    
    // icons
    struct LocalIcon {
        int room_no;
        int x, y;
        int expiry;
        bool operator<(const LocalIcon &other) const { return expiry < other.expiry; }
    };
    priority_queue<LocalIcon> icons;

    // room maps
    struct RoomData {
        // Origin is set to the bottom-left corner if the room is mapped, or null
        // if the room is unmapped.
        // Width, height correspond to the size of gmap.
        MapCoord origin;
        int width, height;

        // Items, Tiles
        struct GfxEntry {
            const Graphic *gfx;
            shared_ptr<const ColourChange> cc;
            MiniMapColour col;
            int depth;
        };
        vector<list<GfxEntry> > gmap;
        list<GfxEntry> & lookupGfx(int x, int y) { return gmap[y*width+x]; }
        bool valid(int x, int y) const { return x>=0 && y>=0 && x<width && y<height; }
        
        // Entities
        EntityMap emap;

        // If "awaiting_copy" is set then the room has not yet been copied into the
        // mini-map.
        bool awaiting_copy;

        // ctor
        RoomData() { }
        void create(int w, int h) {
            awaiting_copy = true; width = w; height = h; gmap.resize(w*h);
        }
    };
    map<int, RoomData> rooms;
    int current_room;
    int my_x, my_y;
    MapDirection my_facing;   // Last direction the knight is known to have moved in.
    bool my_approached;
    
    // mini map
    MiniMap mini_map;
    void copyRoomToMiniMap(RoomData &rd);
    void updateKnightIndicator();

    // menu graphics
    const Graphic * menu_gfx_centre;   // 'arrows' graphic
    const Graphic * menu_gfx_empty;    // 'empty square' graphic
    const Graphic * menu_gfx_highlight;  // 'highlight' graphic.
    
    // controller stuff:
    const Controller & controller;
    const Control * tap_control;
    const Control * menu_control[4];
    bool menu_open;
    bool menu_highlight;
    MapDirection menu_highlight_dir;
    int fire_start_time;
    bool attack_mode;
    enum MenuNullEnum { M_OK, M_CTS, M_NULL };
    MenuNullEnum menu_null;
    MapDirection menu_null_dir;

    // last time the knight was known to be alive, and last known position
    int last_known_time;
    int last_known_x, last_known_y;    // square positions
    
    // messages
    struct Message {
        string message;
        int start_time;
        int stop_time;
    };
    deque<Message> messages;  // sorted by start_time
    vector<string> cts_messages;
    int last_mtime;

    // flash screen
    int flash_screen_start;
    
    // skulls and potion bottle
    const SkullRenderer *skull_renderer;
    const PotionRenderer *potion_renderer;
    int nskulls;
    int health;
    PotionMagic magic;
    bool poison_immun;

    // backpack
    struct BackpackEntry {
        const Graphic *gfx;
        const Graphic *overdraw;
        int no_carried;
        int no_max;
    };
    map<int, BackpackEntry> backpack;

    // gfx buffer.
    map<int, vector<GraphicElement> > gfx_buffer;

    // won/lost status
    bool won;
    bool lost;

    // cheat controls
    const vector<const Control *> *cheat_controls;
};



#endif
