/*
 * find_font_names.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "find_font_names.hpp"
#include "rstream/rstream.hpp"

#include <algorithm>
#include <cctype>
#include <functional>
#include <string>
#include <vector>
using namespace std;

#ifdef USE_FONTCONFIG
#include <fontconfig/fontconfig.h>
#endif

namespace {
    bool initialized = false;
    vector<string> outline_font_names;
    vector<string> bitmap_font_names;
    
    void CookFontNames()
    {
#ifdef USE_FONTCONFIG
        if (outline_font_names.empty()) return; // force use of bmp if they give no outline fonts.

        FcInit();
        
        FcPattern *pattern = FcPatternCreate();
        FcPatternAddInteger(pattern, FC_SLANT, FC_SLANT_ROMAN);
        FcPatternAddInteger(pattern, FC_WEIGHT, FC_WEIGHT_NORMAL);
        FcPatternAddInteger(pattern, FC_WIDTH, FC_WIDTH_NORMAL);
        //FcPatternAddBool(pattern, FC_ANY_METRICS, FcTrue);

        for (vector<string>::iterator it = outline_font_names.begin(); it != outline_font_names.end(); ++it) {
            FcPatternAddString(pattern, FC_FAMILY, (const FcChar8*)(it->c_str()));
        }

        outline_font_names.clear();
        
        FcDefaultSubstitute(pattern);
        FcConfigSubstitute(0, pattern, FcMatchPattern);

        FcResult dummy;
        FcFontSet *matching_set = FcFontSort(0, pattern, false, 0, &dummy);

        if (matching_set) {
            for (int i = 0; i < matching_set->nfont; ++i) {
                FcChar8 * filename = 0;
                FcPatternGetString(matching_set->fonts[i], FC_FILE, 0, &filename);
                if (filename) outline_font_names.push_back((const char*)filename);
            }

            FcFontSetDestroy(matching_set);
        }

        FcPatternDestroy(pattern);
        
#elif defined(WIN32)

        const string font_dir = string(getenv("WINDIR")) + string("\\Fonts\\");
        for (vector<string>::iterator it = outline_font_names.begin(); it != outline_font_names.end(); ++it) {
            *it = font_dir + (*it) + ".ttf";
        }
        
#endif
    }
    
    void Initialize()
    {
        if (initialized) return;
        initialized = true;
        
        RStream str("fonts.txt");
        while (str) {
            string x;
            getline(str, x);

            // Left trim
            x.erase(x.begin(), find_if(x.begin(), x.end(), not1(ptr_fun<int,int>(isspace))));
            // Right trim
            x.erase(find_if(x.rbegin(), x.rend(), not1(ptr_fun<int,int>(isspace))).base(), x.end());

            if (x.empty()) continue;

            if (x[0] == '+') bitmap_font_names.push_back(x.substr(1));
            else if (x[0] != '#') outline_font_names.push_back(x);
        }

        CookFontNames();
    }
}

int GetNumOutlineFonts()
{
    Initialize();
    return outline_font_names.size();
}

int GetNumBitmapFonts()
{
    Initialize();
    return bitmap_font_names.size();
}

const string & GetOutlineFontName(int i)
{
    Initialize();
    return outline_font_names[i];
}

const string & GetBitmapFontName(int i)
{
    Initialize();
    return bitmap_font_names[i];
}
