/*
 * sdl_graphic.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "binary_io/binary_io.hpp"
#include "cfg.hpp"
#include "loadbitmap.hpp"
#include "my_exceptions.hpp"
#include "sdl_graphic.hpp"
#include "sdl_pixels.hpp"
#include "rstream/rstream_rwops.hpp"

// (TODO) see if changing rle_accel flag makes any difference :)
const bool use_rle_accel = true;
const unsigned int sdl_color_key_flag = SDL_SRCCOLORKEY | (use_rle_accel ? SDL_RLEACCEL : 0);

int GetWidth(const Graphic &g) {
	return g.getWidth();
}

int GetHeight(const Graphic &g) {
	return g.getHeight();
}

shared_array<RGB> GetPixels(const Graphic &g) {
    return g.getPixels();
}

int GetID(const Graphic &g) {
	return g.getID();
}

int GetHX(const Graphic &g) {
    return g.getHX();
}

int GetHY(const Graphic &g) {
    return g.getHY();
}

int GetR(const Graphic &g) {
    return g.getR();
}

int GetG(const Graphic &g) {
    return g.getG();
}

int GetB(const Graphic &g) {
    return g.getB();
}

Graphic::Graphic(const string &filename_, shared_array<RGB> pixels_, int width_, int height_,
                 int hx_, int hy_, int r_, int g_, int b_)
    : surface(0), filename(filename_), pixels(pixels_), width(width_), height(height_),
      hx(hx_), hy(hy_), r(r_), g(g_), b(b_), id(0), need_reload(false)
{
	load();
}

Graphic::~Graphic()
{
	unload();
}

void Graphic::load() const
{
	if (!surface) {
		SDL_Surface *temp = SDL_CreateRGBSurface(r>=0 ? SDL_SRCCOLORKEY : 0,
                                                 width, height,
                                                 24,
                                                 0xff0000,
                                                 0xff00,
                                                 0xff,
                                                 0);
        SDL_LockSurface(temp);
        const RGB* rgb = pixels.get();
        for (int y=0; y<temp->h; ++y) {
            for (int x=0; x<temp->w; ++x, ++rgb) {
                PlotPixel(temp, x, y, rgb->r, rgb->g, rgb->b);
            }
        }
        SDL_UnlockSurface(temp);
        if (r >= 0) {
            SDL_SetColorKey(temp, sdl_color_key_flag, SDL_MapRGB(temp->format, r, g, b));
        }

        surface = SDL_DisplayFormat(temp);
        SDL_FreeSurface(temp);
    }
 }

void Graphic::unload() const
{
	if (surface) {
		SDL_FreeSurface(surface);
		surface = 0;
	}
}

void Graphic::blit(SDL_Surface *dest, int sdl_x, int sdl_y) const
{
	// Try to reload if necessary...
	if (need_reload) {
		if (surface) {
			int result = SDL_LockSurface(surface);
			if (result != 0) return;  // can't yet reload
			SDL_UnlockSurface(surface);
			unload();
		}
        load();
        need_reload = false;
	}
	
	// Check if the surface is loaded.
	if (!surface) return;

	// correct for handle position
	sdl_x -= hx;
	sdl_y += hy;

	// Set up dest_rect (SDL ignores the width and height parts, only need to set x and y)
	SDL_Rect dest_rect;
	dest_rect.x = sdl_x;
	dest_rect.y = sdl_y;
	
	// Do the blit, and handle any errors.
	int result = SDL_BlitSurface(surface, 0, dest, &dest_rect);
	if (result == -1) {
		throw UnexpectedError("blit failure");
	} else if (result == -2) {
		// This means that the surface was lost.
		need_reload = true;
	}
}

Graphic * CreateGraphic(shared_array<RGB> pixels, int width, int height, int hx, int hy, int r, int g, int b)
{
	return new Graphic("", pixels, width, height, hx, hy, r, g, b);
}

Graphic * CreateGraphicFromFile(const string &filename, int hx, int hy, int r, int g, int b)
{
    int width, height;
    RStream str(filename);
    shared_array<RGB> pixels = gfx_loadbitmap(str, &width, &height);
    return new Graphic(filename, pixels, width, height, hx, hy, r, g, b);
}


void DeleteGraphic(Graphic *g)
{
	delete g;
}


//
// (binary) serialization
//

void SerializeGraphic(ostream &str, Graphic &g, int id)
{
    if (g.filename.empty()) {
        throw UnexpectedError("graphic serialization failed");
    }
    BinWriteString(str, g.filename);
	BinWriteInt(str, g.hx);
	BinWriteInt(str, g.hy);
	BinWriteInt(str, g.r);
	BinWriteInt(str, g.g);
	BinWriteInt(str, g.b);
	g.id = id;
}

Graphic * CreateGraphicFromStream(istream &str)
{
	string filename = BinReadString(str);
	int hx = BinReadInt(str);
	int hy = BinReadInt(str);
	int r = BinReadInt(str);
	int g = BinReadInt(str);
	int b = BinReadInt(str);
	return CreateGraphicFromFile(filename, hx, hy, r, g, b);
}
