/*
 * sdl_sound_manager.cpp
 *
 * SDL implementation of SoundManager.
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "my_exceptions.hpp"
#include "rstream/rstream_rwops.hpp"
#include "sound_manager.hpp"

#include "SDL.h"

#include <list>
#include <map>
using namespace std;

namespace {
	struct Voice {
		float pos;
		float pmax;
		const Sint16 *data;
		int freq;
	};

	struct Sample {
		Sint16 *data;
		int len;
	};
}

struct SoundManagerImpl {
	bool sound_up;
	SDL_AudioSpec spec;

	list<Voice> voices;
	map<string,Sample> samples;
};

namespace {
	SoundManager *g_sound_manager = 0;

	void FlipEndian(Uint8 *data, int nbytes)
	{
		while (nbytes > 0) {
			swap(*data, *(data+1));
			nbytes -= 2;
		}
	}

	const Sample& LoadSound(const string &name, SoundManagerImpl &impl)
	{
		pair<map<string,Sample>::iterator, bool> p = impl.samples.insert(make_pair(name, Sample()));
		if (p.second) {
			// We have to load this sample

			p.first->second.data = 0;
			p.first->second.len = 0;			

			SDL_AudioSpec spec;
			memset(&spec, 0, sizeof(spec));
			Uint8* data = 0;
			Uint32 len = 0;

			SDL_LoadWAV_RW(RWFromRStream(name), 1, &spec, &data, &len);

			if (spec.format != AUDIO_S16LSB && spec.format != AUDIO_S16MSB) {
				throw GraphicLoadFailed("unsupported sound format: " + name);
			}
			if (spec.format != AUDIO_S16SYS) {
				FlipEndian(data, len);
			}

			if (!data) {
				throw GraphicLoadFailed("failed to load sound file '" + name + "'; SDL error was " + SDL_GetError());
			}

			p.first->second.data = (Sint16*)data;
			p.first->second.len = len/2;   // len is in bytes, p.first->second.len is in (16-bit) words
		}
		return p.first->second;
	}

	
	void my_audio_callback(void *userdata, Uint8 *stream, int len)
	{
		static const float volume_factor = CfgInt("sound_volume") / 100.0f;
		
		Sint16 *buf = reinterpret_cast<Sint16*>(stream);
		SoundManagerImpl *pimpl = static_cast<SoundManagerImpl*>(userdata);
		
		// Zero out the buffer
		memset(buf, 0, len/2);

		// for each voice currently playing
		for (list<Voice>::iterator it = pimpl->voices.begin(); it != pimpl->voices.end(); ) {
			bool erase = false;
			// for each data point
			for (int n=0; n<len/2; ++n) {
				if (it->pos >= it->pmax) {
					erase = true;
					break;
				} else {
					// linearly interpolate
					const int p1 = int(it->pos);
					const float lambda = it->pos - p1;
					const Sint16 x1 = it->data[p1];
					const Sint16 x2 = it->data[p1+1];
					const float interp = (1-lambda)*x1 + lambda*x2;
					buf[n] += Sint16(volume_factor*interp);
					it->pos += (float(it->freq) / pimpl->spec.freq);
				}
			}
			if (erase) {
				list<Voice>::iterator it2 = it++;
				pimpl->voices.erase(it2);
			} else {
				++it;
			}
		}
	}
}

SoundManager& SoundManager::instance()
{
	if (!g_sound_manager) {
		g_sound_manager = new SoundManager;
	}
	return *g_sound_manager;
}

void SoundManager::destroyInstance()
{
	delete g_sound_manager;
	g_sound_manager = 0;
}

SoundManager::SoundManager()
{
	pimpl.reset(new SoundManagerImpl);

	// Init SDL audio
	SDL_AudioSpec desired;
	memset(&desired, 0, sizeof(desired));
	memset(&pimpl->spec, 0, sizeof(pimpl->spec));
	desired.freq = 22050;
	desired.format = AUDIO_S16SYS;
	desired.channels = 1;  // we don't do stereo at the moment.
	desired.samples = 1024;
	desired.callback = my_audio_callback;
	desired.userdata = pimpl.get();

	int err = SDL_OpenAudio(&desired, &pimpl->spec);
	pimpl->sound_up = (err == 0);

	if (pimpl->sound_up) {
		SDL_PauseAudio(0);
	}
}

SoundManager::~SoundManager()
{
	if (pimpl->sound_up) {
		SDL_PauseAudio(1);
		for (map<string,Sample>::iterator it = pimpl->samples.begin();
		it != pimpl->samples.end(); ++it) {
			SDL_FreeWAV((Uint8*)(it->second.data));
		}
		SDL_CloseAudio();
	}
}

void SoundManager::loadSound(const string &name)
{
	LoadSound(name, *pimpl);
}

void SoundManager::playSound(const string &name, int frequency)
{
	if (!pimpl->sound_up) return;

	const Sample& s( LoadSound(name, *pimpl) );
	
	if (s.data) {
		// Drop duplicate sounds, ie if the same sound was requested less than 0.1s ago then we
		// drop it. (This is needed because on the split screen mode sounds will typically get
		// played twice, once for each screen, and this can cause some problems.)
		for (list<Voice>::const_iterator it = pimpl->voices.begin(); it != pimpl->voices.end(); ++it) {
			if (it->data == s.data && it->freq == frequency &&
			it->pos / float(frequency) < 0.1f) {
				return;
			}
		}
		
		// Add a new voice playing this sample
		Voice v;
		v.pos = 0;
		v.pmax = s.len - 2;
		v.data = s.data;
		v.freq = frequency;
		SDL_LockAudio();
		pimpl->voices.push_front(v);
		SDL_UnlockAudio();
	}
}
