/*
 * wce_serial.hpp
 *
 * Serial port streams for CE, implementing packet-based semantics.
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef WCE_SERIAL_HPP
#define WCE_SERIAL_HPP

#include <deque>
#include <iostream>
#include <streambuf>
#include <vector>
using namespace std;

#include <windows.h>


class WceSerialBuf : public streambuf {
public:
	virtual ~WceSerialBuf();

protected:
	// Minimal routines needed for a streambuf implementation are overflow, underflow
	// and sync
	// It is also nice to have showmanyc (so that in_avail will work correctly).
	virtual int_type overflow(int_type meta = traits_type::eof());
	virtual int_type underflow();
	virtual int sync();
	virtual streamsize showmanyc();

private:
	friend class WceSerialStream;
	explicit WceSerialBuf(int port);

	bool tryConnect();
	bool isConnected() { return connected; }

	bool fillInputBuffer();
	bool writeOutputBuffer();
	
private:
	// In each of these deques, the last packet is incomplete while the rest are
	// complete packets. 
	deque<vector<char> > input_buf;  // queued input packets 
	enum PacketState { P_WAITING, P_HALFWAY, P_RECEIVING };
	PacketState packet_state;
	int packet_idx;
	deque<vector<char> > output_buf;  // queued output packets 
	int output_idx;
	
	HANDLE handle;
	bool connected;
	bool half_connected;

	// Update (9/7/06): We now call the CreateFile routine from within a separate thread.
	// This prevents the system from locking up if the serial port cannot be opened etc.
	int port;
	HANDLE thread;
	CRITICAL_SECTION crit_sec;
	bool file_opened_yet;

	static DWORD threadRoutine(void *self);
};



//
// WceSerialStream stream itself
//

class WceSerialStream : public iostream {
public:
	explicit WceSerialStream(const int port) : buf(port) { init(&buf); }

	bool tryConnect() { return buf.tryConnect(); }
	
	// Protocol:
	// ** When stream is first opened it is "disconnected". Call tryConnect() periodically.
	// When it returns TRUE then we are connected.
	// ** Send/receive operations will only work on a "connected" stream.
	// ** flush() must be called periodically, in case output was not fully written on the
	// last flush. Note that flush() always finishes off the current packet.
	
private:
	WceSerialBuf buf;
};

#endif
